/*
 * Misc image convertion routines
 * Copyright (c) 2001, 2002, 2003 Fabrice Bellard.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/**
 * @file imgconvert.c
 * Misc image convertion routines.
 */

/* TODO:
 * - write 'ffimg' program to test all the image related stuff
 * - move all api to slice based system
 * - integrate deinterlacing, postprocessing and scaling in the conversion process
 */

#include "avcodec.h"
#include "dsputil.h"

#ifdef USE_FASTMEMCPY
#include "fastmemcpy.h"
#endif

#ifdef HAVE_MMX
#include "i386/mmx.h"
#endif

#define xglue(x, y) x ## y
#define glue(x, y) xglue(x, y)

#define FF_COLOR_RGB      0 /* RGB color space */
#define FF_COLOR_GRAY     1 /* gray color space */
#define FF_COLOR_YUV      2 /* YUV color space. 16 <= Y <= 235, 16 <= U, V <= 240 */
#define FF_COLOR_YUV_JPEG 3 /* YUV color space. 0 <= Y <= 255, 0 <= U, V <= 255 */

#define FF_PIXEL_PLANAR   0 /* each channel has one component in AVPicture */
#define FF_PIXEL_PACKED   1 /* only one components containing all the channels */
#define FF_PIXEL_PALETTE  2  /* one components containing indexes for a palette */

typedef struct PixFmtInfo {
    const char *name;
    uint8_t nb_channels;     /* number of channels (including alpha) */
    uint8_t color_type;      /* color type (see FF_COLOR_xxx constants) */
    uint8_t pixel_type;      /* pixel storage type (see FF_PIXEL_xxx constants) */
    uint8_t is_alpha : 1;    /* true if alpha can be specified */
    uint8_t x_chroma_shift;  /* X chroma subsampling factor is 2 ^ shift */
    uint8_t y_chroma_shift;  /* Y chroma subsampling factor is 2 ^ shift */
    uint8_t depth;           /* bit depth of the color components */
} PixFmtInfo;

/* this table gives more information about formats */
#ifdef WINCE
static PixFmtInfo pix_fmt_info[PIX_FMT_NB] = {
    /* YUV formats */
    {
        "yuv420p",
        3,
        FF_COLOR_YUV,
        FF_PIXEL_PLANAR,
		0,
        1, 
		1, 
        8
    },
    {
        "yuv422",
        1,
        FF_COLOR_YUV,
        FF_PIXEL_PACKED,
		0,
        1, 
		0,
        8
    },

    /* RGB formats */
    {
        "rgb24",
        3,
        FF_COLOR_RGB,
        FF_PIXEL_PACKED,
		0,
        0, 
		0,
        8
    },
    {
        "bgr24",
        3,
        FF_COLOR_RGB,
        FF_PIXEL_PACKED,
        0,
        0, 
		0,
        8
    },

    {
        "yuv422p",
        3,
        FF_COLOR_YUV,
        FF_PIXEL_PLANAR,
        0,
        1, 
		0, 
        8
    },
    {
        "yuv444p",
        3,
        FF_COLOR_YUV,
        FF_PIXEL_PLANAR,
        0,
        0, 
		0, 
        8
    },

    {
        "rgba32",
        4,
        FF_COLOR_RGB,
        FF_PIXEL_PACKED,
		1,
        0, 
		0,
        8
    },

	{
        "yuv410p",
        3,
        FF_COLOR_YUV,
        FF_PIXEL_PLANAR,
		0,
        2, 
		2,
        8
    },
    {
        "yuv411p",
        3,
        FF_COLOR_YUV,
        FF_PIXEL_PLANAR,
        0,
        2, 
		0,
        8
    },

    {
        "rgb565",
        3,
        FF_COLOR_RGB,
        FF_PIXEL_PACKED,
		0,
        0, 
		0,
        5
    },
    {
        "rgb555",
        4,
        FF_COLOR_RGB,
        FF_PIXEL_PACKED,
		1,
        0, 
		0,
        5
    },

    /* gray / mono formats */
    {
        "gray",
        1,
        FF_COLOR_GRAY,
        FF_PIXEL_PLANAR,
		0,
		0,
		0,
        8,
    },
    {
        "monow",
        1,
        FF_COLOR_GRAY,
        FF_PIXEL_PLANAR,
		0,
		0,
		0,
        1
    },
    {
        "monob",
        1,
        FF_COLOR_GRAY,
        FF_PIXEL_PLANAR,
		0,
		0,
		0,
        1
    },

    /* paletted formats */
    {
        "pal8",
        4,
        FF_COLOR_RGB,
        FF_PIXEL_PALETTE,
		1,
		0,
		0,
        8
    },

    /* JPEG YUV */
    {
        "yuvj420p",
        3,
        FF_COLOR_YUV_JPEG,
        FF_PIXEL_PLANAR,
		0,
        1, 
		1, 
        8
    },
    {
        "yuvj422p",
        3,
        FF_COLOR_YUV_JPEG,
        FF_PIXEL_PLANAR,
		0,
        1, 
		0, 
        8
    },
    {
        "yuvj444p",
        3,
        FF_COLOR_YUV_JPEG,
        FF_PIXEL_PLANAR,
		0,
        0, 
		0, 
        8
    },
};
#else

#endif

void avcodec_get_chroma_sub_sample(int pix_fmt, int *h_shift, int *v_shift)
{
    *h_shift = pix_fmt_info[pix_fmt].x_chroma_shift;
    *v_shift = pix_fmt_info[pix_fmt].y_chroma_shift;
}

const char *avcodec_get_pix_fmt_name(int pix_fmt)
{
    if (pix_fmt < 0 || pix_fmt >= PIX_FMT_NB)
        return "???";
    else
        return pix_fmt_info[pix_fmt].name;
}

#undef FIX
